/*
** Module   : FillIn.java
** Abstract : text field editor/display implementation
**
** Copyright (c) 2005-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SVG 20050609   @21446 Initial version.
** 002 SVG 20050620   @21523 DisplayFormat related classes are redesigned.
**                           Now cursor's position and insert mode are
**                           maintained inside Presentation. This allows
**                           to not pass them by reference into
**                           Presentation edit methods. All
**                           com.goldencode.p2j.ui.chui.dataref package
**                           is obsoleted and removed.
** 003 SVG 20050803   @21994 New activate() method added.
** 004 SIY 20050831   @22501 Refactoring (see @@22394-22409) 
** 005 SIY 20050907   @22634 Fixed use of @inheritDoc tag in javadoc
** 006 SIY 20050909   @22666 Fixed sideLabel handling, fixed format 
**                           setting.
** 007 SIY 20050913   @22837 Removed redundant disableAutoZap attribute,
**                           added support for remaining attributes, added
**                           support for SENSITIVE attribute, implemented
**                           Validateable interface.
** 008 SIY 20051205   @24204 Added clear() method.
** 009 GES 20060115   @23918 Honor right alignment (e.g. numbers).
** 010 SIY 20060131   @24205 Use ThinClient special method instead of 
**                           direct calls to Window.message().
** 011 SIY 20060202   @24239 Support for the TO and NO-LABELS options.
** 012 SIY 20060203   @24249 Added support for force-label flag.
** 013 SIY 20060209   @24465 Cleanups and updates to accommodate changes 
**                           in xxxConfig classes.
** 014 SIY 20060210   @24525 Minor changes caused by changes in 
**                           ComponentConfig.
** 015 SIY 20060222   @24637 Refactored and added methods necessary for 
**                           CHOOSE implementation.
** 016 NVS 20060227   @24786 Key actions KA_* are used instead of virtual
**                           keys VK_* for all editing operations.
** 017 NVS 20060313   @25068 Fixed drawing of uninitialized fill-in.
** 018 SIY 20060316   @25129 Support for header mode.
** 019 NVS 20060322   @25197 Added isModified() method since now it's part
**                           of the DataContainer interface.
** 020 SIY 20060322   @25218 Added setNoLabels() to match new
**                           LabeledWidget interface.
** 021 SIY 20060324   @25282 Fixed access to the AUTO-RETURN attribute, 
**                           added browse specific helpers.
** 022 SIY 20060401   @25312 Now headers are handled in more generic way.
** 023 SIY 20060404   @25359 Initial (stub) support for the text group.
** 024 SIY 20060417   @25594 Do not adjust height if it is explicitly set.
** 025 GES 20060503   @25944 Draw the cursor in the last char position if
**                           the widget is aligned on the right edge of
**                           the frame. Clear the last char entered state
**                           when an action key is used.
** 026 GES 20060503   @25957 Pass a flag to notify the presentation class
**                           about whether editing is active or not. This
**                           allows types to render differently when the
**                           field is or is not enabled.
** 027 GES 20060504   @25963 Left and right arrow keys now shift fields
**                           (change focus like BACK-TAB and TAB 
**                           respectively) when at the boundary of the
**                           field.
** 028 GES 20060504   @25964 Up and Down arrow keys now move to the widget
**                           above or below respectively instead of using
**                           the back-tab and tab approach (which was very
**                           incorrect).
** 029 GES 20060505   @25982 Fixed cursor drawing for right aligned
**                           fields.
** 030 GES 20060505   @25993 Fix for left/right/up/down arrow keys
**                           changing focus in numeric fields.
** 031 GES 20060508   @26014 Ring terminal bell when a formatting error
**                           occurs during editing.
** 032 SIY 20060515   @26152 Added getColumnLabelText().
** 033 GES 20060517   @26204 Additional fixes for cursor drawing and 
**                           proper data display for right aligned
**                           and zero-suppressed numeric fields.
** 034 GES 20060517   @26213 New mechanism to allow forcing left
**                           alignment.
** 035 SIY 20060522   @26426 Support for the displayable value of 
**                           different type of variable.
** 036 SIY 20060613   @27228 Use automatic format assignment not only for
**                           header items.
** 037 SIY 20060616   @27344 Minimize amount of calls to Frame.relayout().
** 038 SIY 20060703   @27747 Added dynamic expansion of character value 
**                           formats.
** 039 NVS 20060630   @27770 Replaced the call to requestFocus() with call
**                           to tryFocusChange() since it should have
**                           Progress semantics.
** 040 GES 20060704   @27798 Removed some unnecessary color propagation 
**                           logic that caused labels (and separator
**                           lines) to be displayed incorrectly when a
**                           non-default color is used in a fill-in.
** 041 NVS 20060703   @27802 getParent.nextFocus() replaced with ThinClient
**                           .nextFocus() call to maintain Progress
**                           semantics. Similar replacement has been done
**                           to getParent.previousFocus().
** 042 SIY 20060705   @27840 Limit minimal format to x(8) for automatic
**                           format assignment. Made sure that all relevant
**                           calculations in setAttrFormat() are always
**                           performed.
** 043 GES 20060710   @27942 moveAboveCursor and moveBelowCursor need to
**                           work using the nearest row, not simply the
**                           immediate row above or below.
** 044 GES 20060711   @27967 Fixed cursor positioning in the default
**                           number formats.
** 045 NVS 20060713   @28026 Direct drawing of the widget has been changed
**                           to comply with new paint/draw approach,
**                           including cursor drawing.
** 046 SIY 20060717   @28080 Added helper methods.
** 047 SIY 20060726   @28229 Removed debugging and unused code. 
** 048 NVS 20060807   @28463 Auto-completion didn't draw the completed part
**                           of the widget. Added drawing for all cases.
** 049 SIY 20060811   @28555 Fixed displaying of enabled but not assigned
**                           widget.
** 050 NVS 20060816   @28613 Optimized drawing by relocating a call to
**                           getLocationOnScreen() out of the loop.
** 051 NVS 20060821   @28705 Fixed a regression in drawing. It's possible
**                           now that the draw() method gets called while
**                           the widget is not placed yet. Such a call is
**                           simply ignored.
** 052 SIY 20060824   @28850 Removed getP2JType() method.
** 053 NVS 20060824   @28873 A number of private methods has been moved to
**                           the FrameHelper class. These methods are used
**                           now in more than one widget (Button).
** 054 SIY 20060825   @28892 Minor cleanup in comments.
** 055 SIY 20060825   @28893 Fixed trap in setValue() when assigning zero
**                           length string to non-string widget. 
** 056 GES 20060914   @29600 Right alignment detection must be aware of
**                           a temporary variable type override. Force
**                           height to 1 if an attempt to set a positive
**                           width and a non-positive height (otherwise
**                           the auto-resize is disabled but the height
**                           is 0 and BAD things happen). Update drawing
**                           code to fully process the temporary override
**                           case (it previously short circuited).
** 057 GES 20060915   @29619 Added format string overrides to the
**                           temporary value override case already in
**                           the setValue() method. Also removed some
**                           format string re-calc that was probably
**                           there for @ base fields but which is now
**                           causing problems.
** 058 GES 20060920   @29725 Fixed multiple problems related to the
**                           processing of unknown value.  In particular,
**                           with an unknown value, the field is always
**                           left aligned.  In addition, unknown literals
**                           passed to setValue() are now honored.
** 059 GES 20060920   @29726 Fixed editing when a temporary override is
**                           active.
** 060 GES 20060920   @29733 Removed extra optional sign space on left
**                           when displaying unknown value in a numeric
**                           field.
** 061 GES 20060920   @29754 setValue() must handle override of number
**                           types as the same type (if the field is a
**                           decimal and there is an integer override,
**                           the integer value is treated as if it was 
**                           a normal assignment rather than an override).
** 062 GES 20060929   @30021 setValue() must handle character type
**                           overrides with wider formats as overrides
**                           instead of bypassing that case as it
**                           artificially caused the data to be truncated.
** 063 SIY 20060925   @30022 Implemented support for TEXT grouping mode.
** 064 NVS 20061005   @30164 getText(boolean) method is package private
**                           now. It's called from Frame to get the screen
**                           value of a widget with overrides included.
** 065 SIY 20061009   @30245 Fixed handling of 1-char words at the end of 
**                           line.
** 066 SIY 20061010   @30254 Added support for DELETE-LINE event. 
** 067 GES 20061010   @30268 Added support for RECALL event.
** 068 SIY 20061122   @31362 Changes for browse-specific insert state
**                           behavior support.
** 069 GES 20061128   @31435 Removed dead code.
** 070 NVS 20061204   @31553 Switched to using a private widgetFocused
**                           flag instead of config.selected
** 071 SIY 20061215   @31699 Fixed handling of CLEAR in group mode.
** 072 SIY 20070108   @31751 Fixed AT field format override.
** 073 NVS 20070426   @33252 Enabled state is checked instead of SENSITIVE
**                           attribute.
** 074 NVS 20070507   @33417 The target widget for the auto-generated GO
**                           event is the containing frame, not this
**                           fill-in.
** 075 SIY 20070801   @34747 Do not touch alignment in redirected mode.
** 076 SIY 20070807   @34815 Added isColumnFixed() method.
** 077 SIY 20070808   @34834 Fixed alignment for numeric formats.
** 078 SIY 20070827   @34925 Numeric fields in headers are also right 
**                           aligned regardless from the actual format
**                           Ref: #34927.
** 079 SIY 20071005   @35383 Label alignment handled separately from data
**                           alignment. 
**                           Ref: #35371.
** 080 NVS 20071011   @35429 call to invalidate() was replaced with the
**                           call to repaint() which has the intended
**                           effect. There are more calls throughout the
**                           file that have to be checked with this info
**                           in mind.
** 081 SIY 20071012   @35442 Fixed alignment of unknown values.
**                           Ref: #35439#1, #35441.
** 082 SIY 20071015   @35472 More fixes in alignment of unknown values.
**                           Ref: #35474
** 083 EVL 20071020   @35492 Restoring the Window insert mode to the state
**                           was used before starting change the numeric
**                           fields. Modification details - #35493.
** 084 EVL 20071107   @35722 Undo changes from #083. The current progress
**                           behavior is exactly the same as p2j before
**                           the fix.
** 085 SVL 20071221   @36426 Label is shown/hidden together with the
**                           widget.
** 086 CA  20080114   @36808 A numeric widget with a side-labeled frame
**                           needs to be left-aligned.
** 087 GES 20080130   @36934 Fix for regression where forceLeft is not
**                           honored. Previously, in right alignment
**                           checking, the forceLeft had lower precedence
**                           than the NumberFormat processing.  In a
**                           side-labeled frame (like tinyInput() which
**                           uses forceLeft) with a number type fill-in,
**                           the result was a right aligned widget. Now
**                           forceLeft is the first thing checked and it
**                           overrides all other alignment calculations.
** 088 CA  20080225   @37194 A decimal value displayed with an integer
**                           field formatted with fractional digits must 
**                           display the actual fractional digits and not  
**                           round the displayed value. Ref: #37193
** 089 SVL 20080226   @37159 draw() is called in completeEdit() if
**                           checkFormat() has not passed - to reflect
**                           changes made by auto-complete. 
** 090 NVS 20080320   @37615 Fixed drawing of fill-ins in focus when
**                           they are disabled.
** 091 NVS 20080401   @37775 Added isFocusable() method implementation.
** 092 SIY 20080401   @37794 Added support for correct event handling when
**                           WAIT-FOR event loop is inactive.
**                           Ref: #37791
** 093 SIY 20080403   @37800 Unknown values are forced to go via 'AUTO-ZAP' 
**                           execution path even if AUTO-ZAP is not active
**                           and widget is not just activated.
** 094 SIY 20080404   @37828 Cursor positioning fixes. 
** 095 SIY 20080409   @37884 Forward UP/DOWN cursor keys to browse when
**                           working from within it.
**                           Ref: #37887
** 096 SIY 20080410   @37899 Send VALUE-CHANGED when content is changed.
**                           Ref: #37882
** 097 NVS 20080420   @38048 processKeyEvent() has been fixed to work in
**                           I18N environment.
** 098 NVS 20080423   @38103 Added support for reimplemented field mode
**                           CHOOSE processing, which consists of new
**                           drawing modes and keyboard input processing.
** 099 EVL 20080428   @38128 Adding activate() call to the focus gaining
**                           tracker. It is required to fix the cases when
**                           the FillIn is activating indirectly. One
**                           example F1 postprocessing.
** 100 SIY 20080521   @38363 Fixed keyboard handling for browse mode.
**                           Ref: #38355
** 101 SIY 20080522   @38367 Added reset() method.
**                           Ref: #38365
** 102 SIY 20080522   @38396 Rewritten reset() method to avoid use of
**                           code which may trigger drawing or cursor 
**                           movement.
** 103 SIY 20080529   @38430 Do not reset static widgets.
** 104 SIY 20080602   @38469 Fixed content scrolling for cases when format
**                           is longer than explicitly set width.
** 105 SIY 20080602   @38504 Fix for #104. In some cases screenValue can 
**                           be null.
** 106 SIY 20080605   @38548 Fix for another #104 regression: when screen
**                           value is shorter than widget width it causes 
**                           misalignment. 
** 107 SIY 20080605   @38574 Avoid repainting where possible.
** 108 SIY 20080623   @38971 FillIn inserted in Browse is always focused.
** 109 NVS 20080704   @39065 Minor optimizations in BACKSPACE handling.
** 110 SVL 20080826   @39576 Skip insertion of a character into the
**                           current row if it has been inserted to the
**                           next row (in group mode).
** 111 SVL 20080914   @39798 Insert and zap modes logic reworked.
** 112 SVL 20080919   @39861 Zap mode logic fixed for some cases.
** 113 SVL 20080922   @39896 Fixed auto-zap behavior when we move from
**                           a non-fill-in widget to a browse cell using
**                           arrows.  
** 114 SIY 20081020   @40169 Added tracking of highlighted drawing.
** 115 SVL 20081022   @40194 Changed insert mode logic for the group mode. 
** 116 CA  20081023   @40164 Added resetCaret - this will reset the cursor
**                           position and will clear the data on the next
**                           key press.
** 117 SIY 20081107   @40371 Mark VALUE-CHANGED as special to simplify 
**                           separation of internally generated events 
**                           and application generated ones.
** 118 CA  20081114   @40443 Changes due to new clipping implementation.
** 119 SVL 20081127   @40714 Insert mode is saved on value redisplaying.
** 120 CA  20081128   @40716 On reset, the widget must not be marked as
**                           modified.
** 121 SVL 20081202   @40745 Fixed insertWord() and getLastWord().  
** 122 NVS 20081205   @40812 Standardized the use of getAndResetBitmap().
** 123 CA  20090116   @41135 activate() calls must be aware of ENTRY event:
**                           if they are generated by a FOCUS_GAINED, then
**                           the justActivated field must not be touched,
**                           as this is not a real PROGRESS entry event.
** 124 CA  20090120   @41176 refill() must take into account current insert
**                           mode - if insert is off, then SPACE key must
**                           not wrap the line to the next one.
** 125 SVL 20090120   @41183 Added solution for the unknown date value
**                           issue.
** 126 CA  20090123   @41203 Fix a regression introduced by #41176: in non-
**                           insert mode, wrapping is allowed when adding a
**                           char for the last word in the line.
** 127 CA  20090123   @41208 Deactivated H123 as it exposes latent issues
**                           for ENTRY event processing. It should be 
**                           reactivated once ENTRY event is fixed.
** 128 SVL 20090123   @41202 Display separators in edit mode if date
**                           fill-in has uninitialized screen buffer.
** 129 SIY 20090203   @41254 Renamed 'highlighted' to 'selected' in order
**                           to avoid name conflict with Component class.
** 130 SVL 20090219   @41336 On recall make auto-zap active and reset
**                           insert mode.
** 131 CA  20090304   @41409 Reactivated H123.
** 132 SVL 20090305   @41391 Insert indicator state can be restored from
**                           the value set on entering/leaving prompting
**                           blocks.
** 133 SVL 20090305   @41395 In the group mode, do not refresh the status 
**                           string when we are moving between two
**                           fill-ins with undefined help string.
** 134 SVL 20090318   @41636 In the group mode, if DEL is pressed, do not
**                           append SPACE char before the merged line, if
**                           current line has no text.
** 135 SVL 20090320   @41659 Do not reset insert mode after transitions
**                           caused by TAB/RETURN/GO key presses if
**                           widgets are situated into different frames.
**                           Also, insert mode should be modified
**                           according to appDataPres mode only if fill-in
**                           has the focus.
** 136 GES 20090424   @41963 Import change.
** 137 SIY 20090508   @42133 Fixed wrapping in group mode. 
** 138 SIY 20090511   @42134 Pressing space bar at last position in group mode 
**                           does not insert space, just moves cursor to the 
**                           next row.
** 139 SVL 20090527   @42473 Enable zap for the case when we have tried to
**                           leave fill-in using TAB, BACK-TAB, GO or RETURN
**                           keys, but failed because no widgets have applied
**                           ENTRY event.
** 140 SIY 20090624   @42890 Fixed cursor positioning for unknown value.
** 141 SIY 20090701   @43021 Reworked format error checks/reporting in order to
**                           support event-based instead of immediate error
**                           reporting.
** 142 SIY 20090709   @43123 Fixed uninitialized date displaying.
** 143 SIY 20090709   @43131 Fixed NPE in cases when FillIn tells browse to
**                           go to the next row and then tries to redraw itself
**                           while there is no parent anymore.
** 144 SIY 20090710   @43133 Fixed uninitialized widget displaying.
** 145 SIY 20090720   @43292 Fixed handling of cursor Up/Down keys.
** 146 SIY 20090906   @43829 Refactored common refreshing code. Don't allow 
**                           drawing of disconnected widget.
** 147 SVL 20091119   @44440 Do not perform requestFocus() if fill-in has no
**                           parent container.
** 148 SIY 20100111   @44523 Fixed a rare case when drawing of empty space
**                           instead of label is necessary.
** 149 EVL 20100406   @44792 Fix for triggering Insert mode indication when
**                           pressing character ?.
** 150 EVL 20100420   @44827 Fix for triggering the Insert mode on when FillIn
**                           is the part of scrollable frame content.
** 151 SIY 20100619          Switching to own implementation of character UI.
** 152 SIY 20100804          Changes in imports, some refactorings.
** 153 SIY 20100822          Refactored AttributeContainer support.
** 154 SIY 20100831          Changes related to refactoring of FrameHelper.
** 155 EVL 20100914          Change to turn zap activation off when in group
**                           mode. It is not clear if this is correct or not
**                           due to missing the real 4GL env. So deep testing
**                           is required to check.
** 156 SIY 20100915          Methods tryFocusChange() and locateFrame() moved 
**                           to UiUtils, minor cleanups.
** 157 SIY 20101122          Switching from KeyEvent to KeyInput.
** 158 SIY 20110112          Changes related to new color handling scheme.
** 159 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 160 SIY 20111202          Refactored and moved from CHUI package.
** 161 AIL 20121206          Update from character to TextOps.
*/

package com.goldencode.p2j.ui.client;


import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager.ErrorEntry;
import com.goldencode.util.*;

/**
 * Implementation of Progress FILL-IN component.
 */
public abstract class FillIn
extends LabeledDataContainer
{
   /** Variable which is bound to this widget. */
   private BaseDataType appVar;
   
   /** Instance of display format. */ 
   protected DisplayFormat appFormat;

   /** Variable which stores our data value at the most recent enable. */
   private BaseDataType backupVar;
   
   /**
    * Internal presentation of displayed data. May be <code>null</code>
    * which means that value can't be represented with the current format.
    * Such value is rendered as several question marks and can only be reset
    * to a valid value by ZAP or CLEAR actions. At resetting corresponding
    * Presentation instance is generated and is assigned to this member.
    */
   private DisplayFormat.Presentation appDataPres;

   /** Alternative variable to display. */
   private BaseDataType displayVar = null;
   
   /** Alternative format to display. */ 
   private DisplayFormat displayFormat = null;
   
   /** Alternative data presentation to display. */
   private DisplayFormat.Presentation displayDataPres = null;
    
   /**
    * Cursor's offset in widget. The left-most position is 0. This member is
    * used only when <code>varDataPres</code> is <code>null</code>.
    */
   private int offsCursor = 0;

   /** Offset of visible area from the beginning of screen value. */
   protected int offsScroll = 0;
   
   /** Stores amount of pad on left side for right aligned widgets. */
   protected int leftPad = 0;
   
   /** Marks a blank (all spaces) right aligned display value. */
   protected boolean cleared = false;

   /** Is field just activated by Tab. */
   protected boolean justActivated = false;

   /** TEXT clause mode state. */
   private boolean groupMode = false;
   
   /** Component attribute container. */
   protected FillInConfig config = null;
   
   /** Use underline attribute if enabled. */
   protected boolean useUnderline = true;

   /** Draw line instead of last line of spaces if height is greater than 1. */
   protected boolean drawLine = false;
   
   /** Width delta for the header mode. */
   private int deltaWidth = 0;

   /** Track if format set explicitly or it is determined automatically. */
   private boolean autoFormat = false;
   
   /** Browse widget to notify. */
   protected Browse browse = null;
      
   /** Flag to force left alignment. */
   private boolean forceLeft = false;

   /** Skip next activation event. */
   private boolean skipActivation = false;
   
   /**
    * Indicates whether we should skip help string adjustment on focus
    * gaining.
    */
   private boolean skipHelpAdjustment = false;
   
   /** Widget in focus flag */
   protected boolean widgetFocused = false;

   /**
    * Indicates whether the key has been applied to the next row (in group
    * mode).
    */
   private boolean keyApplied = false;

   /** Indicates whether zap mode should be reset. */
   private boolean pendingNoZap = false;

   /** Track 'selected' state. */
   protected boolean selected = false;

   /** Hard format check state flag. */
   private boolean formatCheck = false;

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   cfg
    *          Configuration data.
    */
   public FillIn(ComponentConfig cfg)
   {
      config = new FillInConfig(this, cfg);
      config.applyConfig(cfg);

      updateSize();
      addFocusListener(new LocalFocusListener());
      
      // not in edit mode by default
      // use direct field access to avoid repainting
      disable();
   }

   /**
    * Enable/disable the widget for editing.  During this operation, any
    * data values are saved off for subsequent restore (during editing) using
    * the <code>RECALL</code> key.
    *
    * @param    enable
    *           <code>true</code> to enable editing in this widget.
    */
   @Override
   public void setEnabled(boolean enable)
   {
      if (enable)
      {
         BaseDataType bdt = (displayVar == null) ? appVar : displayVar;
         backupVar = (bdt != null) ? bdt.duplicate() : null;

         if (!isEnabled())
            setDrawDateSeparators(true);
      }

      // make sure that we will not trigger repainting preliminarily
      if (enable) 
         enable();
      else
         disable();
      
      if (isVisible())
         repaint();
   }

   /**
    * Let widget to not use underline attribute (for use in
    * <code>Browse</code> widget).
    * 
    * @param   underline
    *          New state of the underlining.
    */
   public void setUnderline(boolean underline)
   {
      useUnderline = underline;
   }
   
   /**
    * Gets the flag determining forced left alignment.
    *
    * @return   <code>true</code> if left alignment is required for this 
    *           field.
    */
   public boolean isForceLeft()
   {
      return forceLeft;
   }
   
   /**
    * Sets the flag to force left alignment.
    *
    * @param    forceLeft
    *           <code>true</code> to force left alignment of this field.
    */
   public void setForceLeft(boolean forceLeft)
   {
      this.forceLeft = forceLeft;
   }
   
   /**
    * Getting the BaseDataType object representing the current value.
    *
    * @return  The value representing the current widget value.
    */
   @Override
   public BaseDataType getValue()
   {
      update();
      return getVariable();
   }

   /**
    * Requesting the focus to the <code>FillIn</code>.
    */
   @Override
   public void requestFocus()
   {
      if (!isEnabled() || parent() == null)
         return;

      //TODO<SIY>: move ThinClient into client package. 
      ThinClient tc = ThinClient.getInstance(); 
      int savedInsState = tc.getSavedInsState();
      int actionCode = tc.getLastActionCode();

      if (savedInsState != -1)
      {
         Window.setInsertMode(savedInsState);
         tc.clearSavedInsState();
      }
      else if (actionCode == Keyboard.KA_TAB ||
               actionCode == Keyboard.KA_BACK_TAB ||
               actionCode == Keyboard.KA_RETURN ||
               actionCode == Keyboard.KA_GO)
      {
         Widget oldFocus = Window.getDefaultWindow().currentFocus();
         
         if (!(groupMode &&
               oldFocus != null &&
               oldFocus instanceof FillIn &&
               ((FillIn) oldFocus).groupMode &&
               oldFocus.isEnabled() &&
               oldFocus.isVisible() &&
               actionCode == Keyboard.KA_RETURN))
         {
            Widget oldFrame = UiUtils.locateFrame(oldFocus);
            Widget newFrame = UiUtils.locateFrame(this);

            if (oldFrame == newFrame && oldFrame != null)
            {
               Window.setInsertMode(Window.INSERT_MODE_OVR);

               // this additionally enables zap for the case when we have
               // tried to leave fill-in using TAB, BACK-TAB, GO or RETURN
               // keys, but failed because no widgets have applied ENTRY event 
               enableZapMode();
            }
         }
      }

      if (appDataPres != null)
      {
         appDataPres.setInsertMode(
            Window.getInsertMode() == Window.INSERT_MODE_INS);
      }
                                                              
      if (browse != null)
      {
         Widget oldFocus = Window.getDefaultWindow().currentFocus();
         pendingNoZap = (oldFocus != null &&
                         !(oldFocus instanceof FillIn) &&
                         (actionCode == Keyboard.KA_CURSOR_RIGHT ||
                          actionCode == Keyboard.KA_CURSOR_LEFT));
      }
      else
      {
         pendingNoZap = (actionCode == Keyboard.KA_CURSOR_UP ||
                         actionCode == Keyboard.KA_CURSOR_LEFT ||
                         actionCode == Keyboard.KA_CURSOR_RIGHT ||
                         actionCode == Keyboard.KA_CURSOR_DOWN);
      }

      super.requestFocus();
   }
   
   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public FillInConfig config()
   {
      return config;
   }
   
   /**
    * The last setup before editing starts. See also
    * {@link com.goldencode.p2j.ui.client.format.DisplayFormat.Presentation#activate() Presentation.activate().}
    * This method is automatically called on <code>focusGained</code> event.
    */
   public void activate()
   {
      activate(true);
   }

   /**
    * The last setup before editing starts. See also
    * {@link com.goldencode.p2j.ui.client.format.DisplayFormat.Presentation#activate() Presentation.activate().}
    * This method is automatically called on <code>focusGained</code> event.
    * 
    * @param    isEntry
    *           Flag indicating that an ENTRY event was generated prior to
    *           this activate call.
    */
   public void activate(boolean isEntry)
   {
      skipHelpAdjustment = false;

      if (skipActivation)
      {
         skipActivation = false;
         return;
      }

      boolean ins = (Window.getInsertMode() == Window.INSERT_MODE_INS);
      
      if (appDataPres != null)
      {
         // Handle special case of @ fields with format override:
         // Upon activation value assigned to @ field is not cleared
         // but original format definition is restored.
         if (displayVar != null && displayDataPres != null)
         {
            BaseDataType save = displayVar;
            clearOverride();
            setValue(save);
         }
         
         appDataPres.activate();
         
         if (browse == null)
            ins = appDataPres.getInsertMode();
      }
      else
      {
         offsCursor = 0;
         
         if (browse == null)
            ins = false;
      }

      if (isEntry)
      {
         int savedInsState = ThinClient.getInstance().getSavedInsState();
         if (savedInsState != -1)
         {
            ins = (savedInsState == Window.INSERT_MODE_INS);
            if (appDataPres != null)
            {
               appDataPres.setInsertMode(ins);
            }
            ThinClient.getInstance().clearSavedInsState();
         }
      }


      if (appDataPres != null && hasFocus())
         setWindowInsertMode(ins);

      if (isEntry)
      {
         // FOCUS_GAINED events must not touch this flag - because will leave
         // the fill-in in an activated state, after a LEAVE trigger returns
         // NO-APPLY and there is a nested UPDATE call which moves focus 
         // away from fill-in, during trigger LEAVE invocation.
         // only ENTRY events can set the fill-in into an activated state.
         justActivated = true;
      }
   }

   /**
    * Fills data structures from variable's value.
    */
   public void display()
   {
      try
      {
         if (appDataPres != null)
         {
            boolean ins = appDataPres.getInsertMode();
            appDataPres = appFormat.fromVar(appVar);
            appDataPres.setInsertMode(ins);
         }
         else
         {
            appDataPres = appFormat.fromVar(appVar);
         }
      }
      catch (DisplayFormatCheckException ex)
      {
         ex.printStackTrace(System.err);
         appDataPres = null;
         ThinClient.getInstance().message(ex.getMessage());
      }

      repaint();
   }

   /**
    * Updates variable's value with current field's content.
    * 
    * @return  Is update successful. <code>true</code> value may also mean
    *          that FillIn contains <b>unsupported </b> (?????) value. It not
    *          an error but variable's values isn't changed.
    */
   public boolean update()
   {
      // null presentation implies that data has been not edited
      if (appDataPres == null)
      {
         return true;
      }

      if (appDataPres.finishEdit())
      {
         // add repainting if required 
      }

      try
      {
         appDataPres.checkFormat();
      }
      catch (DisplayFormatCheckException ex)
      {
         justActivated = true;
         appDataPres.setCursorPos(0);
         return false;
      }

      appDataPres.toVar(appVar);
      return true;
   }

   /**
    * Returns a reference to the variable of this widget.
    * 
    * @return  A reference to the variable of this widget.
    */
   public BaseDataType getVariable()
   {
      return appVar;
   }

   /**
    * Returns format definition string.
    * 
    * @return  Format definition string.
    */
   public String getFormat()
   {
      return appFormat.getFormatDef();
   }

   /**
    * Returns the cursor's offset from the left bound of the FillIn. Progress
    * cursor position equals this offset + 1.
    * 
    * @return  The cursor's offset.
    */
   public int getCursorOffset()
   {
      return (appDataPres != null) ? appDataPres.getCursorPos() : offsCursor;
   }

   /**
    * Returns the last valid cursor position from the left bound of the
    * FillIn.
    * 
    * @return  The last valid cursor offset.
    */
   public int getLastOffset()
   {
      return (appDataPres != null) ? appDataPres.getLastPos() 
                                   : (appFormat.getScreenWidth() - 1);
   }
   
   /**
    * Sets the cursor's offset from the left bound of the FillIn. Progress
    * cursor position equals this offset + 1.
    * 
    * @param   offset
    *          New offset.
    */
   public void setCursorOffset(int offset)
   {
      if (appDataPres != null)
      {
         // take into account leftPad, because drawCaret() will take it into
         // account as well and this results to wrong cursor position during
         // navigation between widgets
         
         if (leftPad != -1)
            offset -= leftPad;
         
         appDataPres.setCursorPos(offset);
      }
      else
      {
         offsCursor = offset;
      }
   }

   /**
    * Resets position of terminal's cursor.
    */
   public void resetCaret()
   {
      appDataPres.moveCursorBegin();
      // next key-press should clear the data, before being applied
      justActivated = true;
      pendingNoZap = false;
   }

   /**
    * Make zap mode enabled, i.e. next key-press will clear the data.
    */
   public void enableZapMode()
   {
      justActivated = true;
      pendingNoZap = false;
   }

   /**
    * Sets position of terminal's cursor which may or may not be the same as
    * the actual cursor location within the widget.  This can be due to 
    * scrolling or due to being a right-edge widget.  All widgets which have 
    * their right-edge up against the right-edge of the frame, will draw the 
    * cursor in the last position of the widget rather than allowing the
    * cursor to move further right.  Note that the drawing position of the
    * cursor does not change the actual recorded cursor position in the
    * class' state.
    */
   @Override
   public abstract void drawCaret();

   /**
    * Draws text. <code>DisplayFormat.Presentation</code> instance is
    * rendered into screen value. Screen value is left justified or padded to
    * the right bound with spaces. Non first lines are filled with spaces. If
    * FillIn is focused then after rendering terminal's cursor is set to
    * appropriate position in the widget.
    */
   @Override
   public void draw()
   {
      selected = false;

      Color dc = UiUtils.getDcolor(this);
      Color pf = config.getPfcolor();

      if (browse != null)
         pf = config.getDcolor();

      Frame frame = (Frame) UiUtils.locateFrame(this);
      boolean inChoose = frame.isInChoose(this);

      if (!screen().isRedirected() && (!isVisible() || !frame.isVisible()))
         return;

      Color color = ((widgetFocused || browse != null) &&
                     isEnabled() && !inChoose) ? pf : dc;

      if (focusTraversable() && useUnderline && !inChoose)
      {
         color = color.withUnderline();
      }

      draw(color);
   }

   /**
    * Actual drawing routine.
    *
    * @param   color
    *          Color for drawing.
    */
   public abstract void draw(Color color);
   
   /**
    * Get screen representation of the data.
    * 
    * @return  Text representation of the variable value according to all
    *          current settings, flags, format, etc.
    */
   @Override
   public String getText()
   {
      return getText(false);
   }
   
   /**
    * Get internal widget alignment policy.
    * 
    * @return  <code>true</code> if widget is formatted for right alignment.
    */
   public boolean isRightAlignedInt()
   {
      if (forceLeft)
         return false;
      
      Frame frame = (Frame) UiUtils.locateFrame(this);

      if (frame != null && getCurrentFormat() instanceof NumberFormat)
      {
         // for header widget force right alignment
         if (config.isHeader())
            return true;
         
         ScrollContainer sc = frame.getContentPane();
         
         // for top labels force right alignment
         if (!sc.isSideLabels())
         {
            return true;
         }
         
         BaseDataType var = (displayVar != null) ? displayVar : appVar;

         if (var != null && var.isUnknown())
            return false;
      }
      
      return getCurrentFormat().isRightPositioning();
   }

   /**
    * Get current active format.
    * 
    * @return  current format.
    */
   private DisplayFormat getCurrentFormat()
   {
      return (displayVar != null && displayFormat != null) ? displayFormat
                                                           : appFormat;
   }
   
   /**
    * Get widget alignment policy.
    * 
    * @return  <code>true</code> if widget is formatted for right alignment.
    */
   @Override
   public boolean isRightAligned()
   {
      if (forceLeft)
         return false;
      
      Frame frame = (Frame) UiUtils.locateFrame(this);
      
      if (frame != null && appFormat instanceof NumberFormat)
      {
         // for header widget force right alignment
         if (config.isHeader())
            return true;
         
         ScrollContainer sc = frame.getContentPane();
         
         // for top labels force right alignment
         if (!sc.isSideLabels())
            return true;
      }

      return appFormat.isRightPositioning();
   }

   /**
    * Process KeyEvents for this widget. Most actual work is performed by
    * <code>DisplayFormat</code> and <code>DisplayFormat.Presentation</code>
    * instances. The latter instance may be absent but in this case only ZAP
    * and CLEAR actions are possible which will generate new
    * <code>DisplayFormat.Presentation</code> instance.
    * <p>
    * Here is the list of the handled keys and the corresponding actions:
    * <ul>
    *    <li><code>KA_BACKSPACE</code>. Erases one symbol before
    *        the cursor position via <code>deleteBack()</code> method;
    *    <li><code>KA_BACK_TAB</code>. Requests container to switch
    *        input focus to the previous widget in the tab sequence;
    *    <li><code>KA_CLEAR</code>. Calls
    *        <code>DisplayFormat.clear()</code> method to make new 
    *        presentation;
    *    <li><code>KA_CURSOR_LEFT</code>, <code>KA_CURSOR_RIGHT</code>.
    *        Moves cursor by one position using <code>moveCursorLeft()</code>
    *        and <code>moveCursorRight()</code> methods;
    *    <li><code>KA_CURSOR_UP</code>, <code>KA_CURSOR_DOWN</code>.
    *        Moves cursor one line or down 
    *    <li><code>KA_DELETE_CHARACTER</code>. Erases one symbol at the cursor
    *        position via <code>deleteCurrent()</code> method;
    *    <li><code>KA_HOME</code>
    *        Moves cursor to the bound of FillIn using <code>moveCursorBegin()
    *        </code> 
    *    <li><code>KA_INSERT_MODE</code>.
    *        Switches insert mode through {@link Window}class;
    *    <li>Printable character. ZAP performed if appropriate. Depending on
    *        current insert mode <code>insert()</code> or 
    *        <code>overwrite()</code> is called after this;
    *    <li><code>KA_RECALL</code>
    *        Restores the original contents of the FillIn; 
    *    <li><code>KA_RETURN</code> and <code>KA_TAB</code>.
    *        Requests container to switch input focus to the next widget in 
    *        the tab sequence;
    * </ul>
    * 
    * @param   ke
    *          Key event structure.
    */
   @Override
   public void processKeyEvent(KeyInput ke)
   {
      Frame frame = (Frame) UiUtils.locateFrame(this);

      if (frame.isInChoose(this))
      {
         frame.processKeyEvent(ke);
         return;
      }
      
      // First call all KeyListener objects that may have been registered
      // for this component.
      super.processKeyEvent(ke);

      // Check if any of the KeyListeners consumed the KeyEvent.
      if (ke.isConsumed())
      {
         return;
      }

      int key = ke.actionCode();

      boolean inputRes = true;
      boolean zapReset = false;
      keyApplied       = false;

      if (key >= 0x20 && key <= 0xff) // range for printable characters
      {
         if (config.isReadOnly())
         {
            return;
         }

         config.setModified(true);

         // honor any Progress triggers
         KeyInput vchng = new KeyInput(this,
                                       EventType.KEY_PRESSED,
                                       Keyboard.SE_VALUE_CHANGED,
                                       true);
         EventManager.postEvent(vchng);
         
         // tracks if we were already at the last char of the field on entry
         // to this method
         boolean pastLast = false;
         boolean canSwitch = true;
         
         // status of the WAIT-FOR loop
         boolean focused = hasFocus();

         char ch = Utils.toChar(key);

         if (focused &&
             (isZapActive() ||
              appDataPres == null ||
              appDataPres.isUnknown()))
         {
            DisplayFormat.Presentation zapDataPres =
               appFormat.clear(appDataPres);
               
            pastLast = zapDataPres.isLastCharEntered();

            inputRes = zapDataPres.input(ch);
            
            // mode is changed even if input fails

            setInsertMode(zapDataPres.getInsertMode());
            
            if (inputRes)
            {
               appDataPres = zapDataPres;
               clearOverride();
            }
            else
            {
               drawCaret();   // return cursor to this widget
               requestSync(); // display possibly shows new insert mode
               return;        // exit with justActivated == true
            }
         }
         else
         {
            // when inserting - refill before inserting character
            
            if (groupMode)
               canSwitch = refill(true, key);

            if (!keyApplied)
            {
               if (focused)
               {
                  pastLast = appDataPres.isLastCharEntered();
                  inputRes = appDataPres.input(ch);
               }
               else
               {
                  // apply char without moving cursor
                  if (isRightAlignedInt())
                     appDataPres.moveCursorEnd();
                  else
                     appDataPres.moveCursorBegin();
               
                  boolean ins = appDataPres.getInsertMode();
               
                  appDataPres.setInsertMode(Window.getInsertMode() ==
                                            Window.INSERT_MODE_INS);
               
                  appDataPres.input(ch);

                  appDataPres.setInsertMode(ins);
               
                  if (isRightAlignedInt())
                     appDataPres.moveCursorEnd();
                  else
                     appDataPres.moveCursorBegin();
               
                  // widget is not in focus, so just repaint it
                  repaint();
               }
            }

            if (!focused)
               return;
         }

         zapReset = true;

         if (!keyApplied && appDataPres.isLastCharEntered())
         {
            if ((config.isAutoReturn() && getDataEntryReturn()) || 
                (groupMode && canSwitch))
            {
               appDataPres.setLastCharEntered(false);
               if (completeEdit())
               {
                  if (browse != null)
                  {
                     // the following call should be conditional
                     ThinClient.nextFocus(this);
                  }
                  else
                  {
                     if (frame.isLastWidget(this))
                     {
                        EventManager.postEvent(new KeyInput(frame,
                                                       EventType.KEY_TYPED, 
                                                       Keyboard.KA_GO));
                     }
                     else
                     {
                        // the following call should be conditional
                        
                        if (groupMode && !isLastInGroup())
                        {
                           FillIn ptr = getNextInGroup();
                           
                           ptr.internalActivation(this);
                           FocusManager.tryFocusChange(ptr);
                        }
                        else
                           ThinClient.nextFocus(this);
                     }
                  }
                  refresh();
                  return;
               }
            }
            else if (pastLast)
            {
               screen().beep();
            }
         }
      }
      else
      {
         // key actions
      
         // clear the last char flag since an edit can only be auto-returned
         // by a data key, not an action
         if (appDataPres != null)
         {
            appDataPres.setLastCharEntered(false);
         }

         // LEAVE event
         if (key == Keyboard.SE_LEAVE)
         {
            // "hard" format check

            ErrorEntry res = null;

            if (isFormatCheck() && appDataPres != null)            
               res = appDataPres.validateFormat();
            
            if (res != null && res.num > 0)
            {
               // force NO-APPLY
               ke.consume();
               
               // inform user about error
               screen().beep();
               ErrorManager.displayError(res.num, res.text);
               // re-activate widget
               activate();

               // condition can be thrown by displayError(), 
               // but we just pass it trough
            }
            
            return;
         }
         
         // RETURN action key 
         if (key == Keyboard.KA_RETURN)
         {
            if (!completeEdit())
               return;
   
            if (getDataEntryReturn())
            {
               if (frame.isLastWidget(this))
               {
                  EventManager.postEvent(new KeyInput(frame,
                                                 EventType.KEY_TYPED,
                                                 Keyboard.KA_GO));
               }
               else
               {
                  // the following call should be conditional
                  
                  if (groupMode && !isLastInGroup())
                  {
                     FillIn ptr = getNextInGroup();
                     
                     if (getInsertMode())
                     {
                        String text = getText();
                        String word = text.substring(getCursorOffset());
                        text = text.substring(0, getCursorOffset());
                        
                        text = StringHelper.safeTrimTrailing(text);
                        word = StringHelper.safeTrimTrailing(word);
                        
                        if (word.length() > 0)
                        {
                           ptr.insertWord(word, false);
                           setValueInt(new character(text));
                        }
                     }

                     setCursorOffset(0);
                     ptr.internalActivation(this);
                     
                     FocusManager.tryFocusChange(ptr);
                  }
                  else
                     ThinClient.nextFocus(this);
               }
               
               refresh();
            }
            return;
         }
               
         // INSERT-MODE action key 
         if (key == Keyboard.KA_INSERT_MODE)
         {
            if (appDataPres != null && !appDataPres.isUnknown())
            {
               justActivated = false;
            }
            pendingNoZap = true;

            setInsertMode(!getInsertMode());
            drawCaret();
            requestSync();
            return;
         }
               
         // TAB action key 
         if (key == Keyboard.KA_TAB)
         {
            if (!completeEdit())
               return;

            if (browse != null)
            {
               browse.processKeyEvent(ke);
               return;
            }
            
            // the following call should be conditional
            
            if (groupMode)
               FocusManager.tryFocusChange(getNextAfterGroup());
            else
               ThinClient.nextFocus(this);

            refresh();
            return;
         }

         if (key == Keyboard.KA_CURSOR_DOWN)
         {
            if (browse != null)
            {
               browse.processKeyEvent(ke);
               return;
            }
            
            if (!completeEdit())
            {
               return;
            }
            
            frame.moveBelowWidget(this, getValidCursorPos());

            return;
         }

         // BACK-TAB action key 
         if (key == Keyboard.KA_BACK_TAB)
         {
            if (!completeEdit())
               return;

            if (browse != null)
            {
               browse.processKeyEvent(ke);
               return;
            }
            
            // the following call should be conditional
            
            if (groupMode)
               FocusManager.tryFocusChange(getPrevBeforeGroup());
            else
               ThinClient.previousFocus(this);

            refresh();
            return;
         }

         // CURSOR-UP action key
         if (key == Keyboard.KA_CURSOR_UP)
         {
            if (browse != null)
            {
               browse.processKeyEvent(ke);
               return;
            }
            
            if (!completeEdit())
            {
               return;
            }
            
            frame.moveAboveWidget(this, getValidCursorPos());

            return;
         }
         
         if (key == Keyboard.KA_DELETE_LINE)
         {
            if (config.isReadOnly() || !groupMode)
            {
               return;
            }
            
            setCursorOffset(0);
            setValueInt(new character(""));
            key = Keyboard.KA_DELETE_CHARACTER;
         }
         
         // CLEAR action key
         if (key == Keyboard.KA_CLEAR)
         {
            if (config.isReadOnly())
            {
               return;
            }

            config.setModified(true);
            
            appDataPres = appFormat.clear(appDataPres);
            
            setFormatCheck(true);
            inputRes = true;
            zapReset = true;
            
            if (groupMode && !isLastInGroup())
            {
               FillIn fin = getNextInGroup();
               
               ScreenBitmap bitmap = screen().getAndResetBitmap();

               while (fin != null)
               {
                  fin.processKeyEvent(ke);
                  
                  // simulate LEFT-END...
                  if (fin.appDataPres == null)
                     fin.offsCursor = 0;
                  else
                     fin.appDataPres.moveCursorBegin();
                  
                  // simulate CLEAR
                  fin.config.setModified(true);
                  fin.appDataPres = fin.appFormat.clear(fin.appDataPres);
                  
                  // refresh screen
                  fin.refresh();
                  
                  fin = fin.getNextInGroup();
               }
               
               screen().setBitmap(bitmap);
            }
         }
         // CURSOR-LEFT action key
         else if (key == Keyboard.KA_CURSOR_LEFT)
         {
            if (appDataPres == null)
            {
               if (offsCursor > 0)
               {
                  --offsCursor;
                  inputRes = true;
               }
               else
               {  
                  // the left most widget ignores this condition
                  if (browse == null && !frame.isLeftMostWidget(this))
                  {
                     // act like a back-tab
                     // the following call should be conditional
                     ThinClient.previousFocus(this);
                  }
               }
            }
            else
            {
               // at the left end already?
               if (appDataPres.getCursorPos() == 0)
               {
                  // the left most widget ignores this condition
                  if (browse == null && !frame.isLeftMostWidget(this))
                  {
                     // act like a back-tab
                     if (!completeEdit())
                     {
                        return;
                     }
                     // the following call should be conditional
                     ThinClient.previousFocus(this);
                     refresh();
                  }
               }
               else
               {
                  // move left 1 position
                  inputRes = appDataPres.moveCursorLeft();
                  zapReset = true;
               }
            }
         }
         // CURSOR-RIGHT action key
         else if (key == Keyboard.KA_CURSOR_RIGHT)
         {
            int end = getLastOffset();
            int cur = getCursorOffset();
            
            if (appDataPres == null)
            {
               if (cur < end)
               {
                  offsCursor++;
                  inputRes = true;
               }
               else
               {
                  // the right most widget ignores this condition
                  if (browse == null && !frame.isRightMostWidget(this))
                  {
                     // act like a tab
                     // the following call should be conditional
                     ThinClient.nextFocus(this);
                  }
               }
            }
            else
            {
               // at the right end already?
               if (cur >= end)
               {
                  // the right most widget ignores this condition
                  if (browse == null && !frame.isRightMostWidget(this))
                  {
                     // act like a tab
                     if (!completeEdit())
                     {
                        return;
                     }
                     // the following call should be conditional
                     ThinClient.nextFocus(this);
                     refresh();
                  }
               }
               else
               {
                  // move right 1 position
                  inputRes = appDataPres.moveCursorRight();
                  zapReset = true;
               }
            }
         }

         // HOME action key
         else if (key == Keyboard.KA_HOME || key == Keyboard.SE_ENTRY)
         {
            // only ENTRY events can set this widget as activated
            activate(true);
         }
         
         // LEFT-END action key
         else if (key == Keyboard.KA_LEFT_END)
         {
            if (appDataPres == null)
            {
               offsCursor = 0;
               inputRes = true;
            }
            else
            {
               inputRes = appDataPres.moveCursorBegin();
               zapReset = true;
            }
         }
         
         // RIGHT-END action key
         else if (key == Keyboard.KA_RIGHT_END)
         {
            if (appDataPres == null)
            {
               offsCursor = appFormat.getScreenWidth() - 1;
               inputRes = true;
            }
            else
            {
               inputRes = appDataPres.moveCursorEnd();
               zapReset = true;
            }
         }
               
         // BACKSPACE action key
         else if (key == Keyboard.KA_BACKSPACE)
         {
            if (appDataPres == null)
            {
               if (offsCursor > 0)
               {
                  --offsCursor;
                  inputRes = true;
               }
            }
            else
            {
               if (config.isReadOnly())
               {
                  return;
               }
               
               if (getCursorOffset() == 0 && groupMode && !isFirstInGroup())
               {
                  // move cursor to previous widget in group and 
                  // forward event to that widget

                  FillIn ptr = getPrevInGroup();
                  
                  ptr.internalActivation(this);
                  
                  if (ptr.appDataPres == null)
                     ptr.offsCursor = ptr.appFormat.getScreenWidth() - 1;
                  else
                     ptr.appDataPres.moveCursorEnd();
                  
                  setCursorOffset(ptr.getCursorOffset());
                  
                  if (FocusManager.tryFocusChange(ptr))
                  {
                     EventManager.postEvent(new KeyInput(ptr,
                                                    EventType.KEY_TYPED, 
                                                    Keyboard.KA_DELETE_CHARACTER));
                  }
               }
               else
               {
                  char chr = ' ';
                  int co = getCursorOffset();                   

                  if (co > 0)
                  {
                     chr = getText().charAt(co - 1);
                  }
                  else if (co == 0)
                  {
                     chr = getText().charAt(0);
                  }
                  
                  config.setModified(true);
                  inputRes = appDataPres.deleteBack();
                  
                  if (groupMode)
                     refill(false, chr);
               }
            }
         }
               
         // DELETE-CHARACTER action key
         else if (key == Keyboard.KA_DELETE_CHARACTER)
         {
            if (appDataPres == null)
            {
               return;
            }

            if (config.isReadOnly())
            {
               return;
            }
            
            char chr = ' ';
            
            if (getCursorOffset() < appFormat.getScreenWidth())
               chr = getText().charAt(getCursorOffset());
            
            config.setModified(true);
            inputRes = appDataPres.deleteCurrent();
            
            if (groupMode)
               refill(false, chr);
         }

         // RECALL action key 
         else if (key == Keyboard.KA_RECALL)
         {
            // restore the value to the state we had at the last call to
            // enable
            if (backupVar != null)
            {
               setValue(backupVar);
               if (appDataPres != null)
               {
                  appDataPres.setInsertMode(false);
               }
               justActivated = true;
               pendingNoZap = false;
               inputRes = true;
            }
         }
                           
         // ignore all unknown actions
         else
            return;
      }

      if (appDataPres != null && hasFocus())
      {
         setWindowInsertMode(appDataPres.getInsertMode());
      }
      
      if (inputRes)
      {
         if (zapReset)
         {
            justActivated = false;
         }
         // direct drawing of the widget
         ScreenBitmap bitmap = screen().getAndResetBitmap();
         refresh();
         screen().setBitmap(bitmap);
      }
      else if (widgetFocused)
      {
         drawCaret();
         requestSync();
      }
   }

   /**
    * Refresh widget.
    */
   private void refresh()
   {
      if (parent() == null || UiUtils.locateFrame(this) == null)
         return;

      draw();
      requestSync();
   }

   /**
    * Clear input variable.
    */
   public void clear()
   {
      if (config.isHeader())
         return;
      
      displayVar = null;
      
      processKeyEvent(new KeyInput(this, 
                                   EventType.KEY_TYPED, Keyboard.KA_HOME));
      processKeyEvent(new KeyInput(this, 
                                   EventType.KEY_TYPED, Keyboard.KA_CLEAR));
   }

   /**
    * Reset widget content.
    */
   public void reset()
   {
      if (config.isStatic())
         return;
      
      if (config.getDataType().equalsIgnoreCase("character") || 
          config.getDataType().equalsIgnoreCase("decimal")   || 
          config.getDataType().equalsIgnoreCase("integer"))
      {
         if (appDataPres != null)
         {
            if (displayVar != null && displayDataPres != null)
            {
               BaseDataType save = displayVar;
               clearOverride();
               setValue(save);
            }

            appDataPres.activate();
         }
         else
            offsCursor = 0;

         appDataPres = appFormat.clear(appDataPres);
         config.setModified(false);
      }
   }
   
   /**
    * Main worker routine for the <code>FillIn</code> text accessor.
    * 
    * @param   isInternal
    *          If this flag is set to <code>true</code> then alternative
    *          variable value is used when available.
    *          
    * @return  Text representation of the variable value according to all
    *          current settings, flags, format, etc.
    */
   public String getText(boolean isInternal)
   {
      DisplayFormat.Presentation curPres = appDataPres;
      DisplayFormat              curFmt  = appFormat;
      
      if (isInternal && displayVar != null && displayDataPres != null)
      {
         curPres = displayDataPres;
         curFmt  = displayFormat;
      }

      String screenValue = (curPres != null && !config.isBlank())
                           ? curPres.toScreenValue(isEnabled())
                           : null;

      // cursor position must be in visible part of field but only if it
      // resides in format bounds
      int fieldCursorPos = Math.min(getCursorOffset(),
                                    curFmt.getScreenWidth() - 1);

      if (fieldCursorPos < offsScroll)
      {
         offsScroll = fieldCursorPos;
      }
      else if (fieldCursorPos >= offsScroll + width())
      {
         offsScroll = fieldCursorPos - width() + 1;
      }

      String display = null;
      
      if (config.isBlank())
      {
         display = DisplayFormat.generateSpaces(width());
      }
      else
      {
         boolean leftTrim = (widgetFocused)
                            ? false 
                            : curFmt.isRightPositioning();
         boolean rightAlign = isRightAlignedInt();
         int     width      = width();
         
         // this section of code implements some fixups and saves the result 
         // of some calculations to properly support the display and cursor
         // positioning of NUMBERS (other data types will all bypass this
         // code below until the formatScreenValue call)
         
         // reset left padding
         leftPad = -1;
         
         // check for special case where the left sign is optional
         if (curPres != null && curPres.isOptionalLeftSign())
         {
            // special case where the optional sign space must be put back
            // because the rendering process removes it (this affects formats
            // such as "-999" which will display 1 char too far left in the
            // non-negative case
            if (!curPres.isUnknown()      &&
                !rightAlign               &&
                !curPres.isNegative()     && 
                curPres.isFixedNumeric())
            {
               StringBuilder sb = new StringBuilder(" ").append(screenValue);
               screenValue = sb.toString();
               
               // we aren't right aligned, but in this case we do need to
               // shift the cursor
               leftPad = 1;
            }
         }
         
         cleared = false;
         
         // calculate the amount of left padding will be present in the case
         // of a right aligned widget
         if (rightAlign)
         {
            int len = screenValue.length();
            DisplayFormat fmt = getCurrentFormat(); 
            
            if (fmt instanceof NumberFormat &&
                screenValue.equals("?")     &&
                fmt.isRightPositioning())
            {
               // initial number formatting case, suppress padding
               len = width;
            }
            
            // calc and save the amount of padding on the left that will 
            // occur (it would be better to get this from DisplayFormat 
            // since we essentially are hard coding the algorithm in 2
            // places, but that would require a more significant change to
            // the interface, so we cheap it and do it here)
            leftPad = (width <= len) ? 0 : (width - len);
            
            cleared = true;
            
            // detect all-blank output and reset the cleared flag in that
            // case
            for (int j = 0; j < len; j++)
            {
               // exit at first non-blank
               if (screenValue.charAt(j) != ' ')
               {
                  cleared = false;
                  break;
               }
            }
         }
         
         if (curFmt instanceof NumberFormat)
         {
            NumberFormat nf = (NumberFormat) curFmt;
            
            if (!rightAlign && 
                nf.isRightPositioning() == rightAlign &&
                nf.isSplitFormat())
            {
               // handle the case when an left-aligned numeric field with a
               // side-labeled frame gets right-aligned
               int savedLen = screenValue.length();
               
               screenValue = screenValue.trim();
               
               int len = screenValue.length();
               int spaceCount = savedLen - len;
               
               StringBuilder sb = new StringBuilder(screenValue);

               // add spaceCount spaces to the end of the screen
               for (int j = 0; j < spaceCount; j++)
               {
                  sb.append(' ');
               }
               screenValue = sb.toString();
            }
         }
         
         if (screenValue != null && 
             screenValue.length() >= width && 
             (offsScroll + width > screenValue.length()))
         {
            StringBuilder sb = new StringBuilder(screenValue);

            int spaceCount = (offsScroll + width) - screenValue.length();
            
            for (int j = 0; j < spaceCount; j++)
            {
               sb.append(' ');
            }
            
            screenValue = sb.toString();
         }

         display = DisplayFormat.formatScreenValue(screenValue,
                                                   width,
                                                   offsScroll,
                                                   leftTrim,
                                                   rightAlign);
      }
      return display;
   }

   /**
    * Changes the insert mode state in the window and the data presentation.
    *
    * @param   mode
    *          New state.
    *          <ul>
    *             <li><code>false</code>- overwrite mode,
    *             <li><code>true</code>- insert mode.
    *          </ul>
    */
   public void setInsertMode(boolean mode)
   {
      if (appDataPres != null)
      {
         appDataPres.setInsertMode(mode);
      }

      setWindowInsertMode(mode);
   }

   /**
    * Check if this instance is a right-edge component in the parent
    * container.
    * 
    * @return  <code>true</code> if this component is a right-edge component
    *          in the frame.
    */
   protected boolean isRightEdgeWidget()
   {
      return (location().x + width()) == parent().width(); 
   }
   
   /**
    * Check if this instance is a left-edge component in the parent
    * container.
    * 
    * @return  <code>true</code> if this component is a left-edge component
    *          in the frame.
    */
   protected boolean isLeftEdgeWidget()
   {
      return (location().x == 0); 
   }
   
   /**
    * Convenience method to get access to SESSION:DATA-ENTRY-RETURN attribute.
    * 
    * @return  Value of the SESSION:DATA-ENTRY-RETURN attribute.
    */
   private boolean getDataEntryReturn()
   {
      WindowConfig cfg = (WindowConfig) Window.getDefaultWindow().config();
      return cfg.getDataEntryReturn();
   }

   /**
    * Sets new widget's size. Resets AUTO-RESIZE attribute.
    * 
    * @param   width
    *          The width component of the new size.
    * @param   height
    *          The height component of the new size.  Will be forced to 1
    *          if <code>width</code> is positive and this value is less than
    *          or equal to 0.
    */
   @Override
   public void setSize(int width, int height)
   {
      config.setAutoResize(false);
      
      if (width > 0 && height <= 0)
      {
         height = 1;
         config.setHeightCharsDirect(height);
      }
      
      super.setSize(width, height);
   }

   /**
    * This component will receive an input focus when Tab or Shift-Tab is
    * pressed.
    * 
    * @return  <code>true</code> if this component is enabled and visible.
    */
   @Override
   public boolean focusTraversable()
   {
      return isEnabled() && isVisible();
   }

   /**
    * Returns this component directly because it is not an usual container and
    * have no child components.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#currentFocus()
    */
   @Override
   public Widget currentFocus()
   {
      return this;
   }

   /**
    * Returns this component directly because it is not an usual container and
    * have no child components.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#defaultFocus()
    */
   @Override
   public Widget defaultFocus()
   {
      return this;
   }

   /**
    * Sets new display format. A current screen value is discarded and widget
    * is refilled from the bound variable. The width may be changed if the
    * size doesn't fixed.
    * 
    * @param   format
    *          New display format definition string.
    * 
    * @return  <code>true</code> if format was changed and <code>false</code>
    *          otherwise.
    *          
    * @throws  DisplayFormatParsingException
    *          Unrecognized variable type or invalid format string.
    */
   public boolean setAttrFormat(String format)
   throws DisplayFormatParsingException
   {
      autoFormat = (format == null);

      if (autoFormat && appVar instanceof character)
      {
         int len = TextOps.lengthOf((character) appVar);

         if (len < 8)
            len = 8;
         
         format = "x(" + len  + ")";
      }

      if (format == null)
         appFormat = DisplayFormatFactory.makeDefaultFormat(appVar);
      else
         appFormat = DisplayFormatFactory.makeFormat(appVar, format);

      updateSize();
      display();

      if (config.getFormat() != null &&
          format != null &&
          format.equals(config.getFormat()))
         return false;
      
      config.setFormatDirect(appFormat.getFormatDef());
      return true;
   }

   /**
    * Sets the new variable value in the fill-in.
    * 
    * @param   value
    *          New value for the screen buffer.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      setValue(value, null);
   }
   
   /**
    * Sets the new variable value in the fill-in.
    * 
    * @param    value
    *           New value for the screen buffer.  May be a temporary override
    *           of the value with differently formatted data (it may even
    *           be of a different data type).
    * @param    fmt
    *           An optional format string to use with this value.  This will
    *           temporarily override the current format string.  If specified
    *           this will also cause the <code>value</code> to be detected as
    *           a temporary override (even if the data type is the same as
    *           the "permanent" data type). If <code>null</code>, then no
    *           format string change is made and a temporary override of the 
    *           <code>value</code> will only be detected if the data types
    *           differ.
    */
   public void setValue(BaseDataType value, String fmt)
   {
      if (value == null)
         return;
      
      // honor the override by the unknown value literal
      if (value instanceof unknown)
      {
         if (appVar != null)
         {
            appVar.setUnknown();
         }
         
         // bypass the override case
         value = appVar;
         fmt   = null;
      }
      
      String  cls   = value.getTypeName();
      String  dtcls = config.getDataType();
      boolean same  = cls.equalsIgnoreCase(dtcls);
      
      // data type fixups/conversion where possible
      if (!same && fmt == null)
      {
         // treat integer and decimal as equivalent
         
         if (cls.equalsIgnoreCase("decimal") && 
             dtcls.equalsIgnoreCase("integer"))
         {
            int fractDigits =  NumberType.getDigitsRight(getFormat()).length(); 

            // if the displayed value is a decimal value and the format has
            // fractional digits, then the value must not be converted to int
            // (the fractional digits must be displayed).
            if (fractDigits == 0) 
            {
               value = new integer((decimal) value);
            }
            else
            {
               value = new decimal((decimal) value);
            }
            same = true;
         }
         
         if (cls.equalsIgnoreCase("integer") && 
             dtcls.equalsIgnoreCase("decimal"))
         {
            value = new decimal((integer) value);
            same = true;
         }
         
         if (same)
         {
            // reset the type name
            cls = value.getTypeName();
         }
      }
      
      // avoid the temporary override where possible
      if (same && fmt != null)
      {
         String master = getFormat();
         
         // check if the format strings are the same
         if (fmt.equals(master))
         {
            // bypass the override case
            fmt = null;
         }
         else
         {
            // non-character types don't have the extended field-sharing
            // behavior, they all respond as straight overrides
            if (cls.equals("CHARACTER"))
            {
               // extended field sharing behavior
               
               // because @ base field usage with character strings often
               // causes an explicit format string to be passed down BUT the
               // data can also be edited, we have to assign that data back
               // to the non-override data variable
               // TODO: investigate if "incompatible" override format strings
               //       will cause this reset to give different results than
               //       the 4GL (e.g. someone passes "!(2)" and the original
               //       field format is "x(5)" or "9(3)"
               appVar = character.valueOf(value, fmt);
               
               // if the new format string is larger than the original, then
               // we need to still treat it as an override since it will be
               // used for display purposes NOW but on any edit only the
               // original format will be honored 
               int mlen = character.preprocessFormatString(master).length();
               int flen = character.preprocessFormatString(fmt).length();
               
               if (mlen >= flen)
               {
                  // since we have pre-truncated the data as needed by the
                  // format, we can bypass the override case
                  fmt = null;
               }
            }
         }
      }
      
      // process the temporary override case
      if (!same || fmt != null)
      {
         displayVar = value.duplicate();
         
         try
         {
            if (fmt == null && value instanceof character)
            {
               int len = TextOps.lengthOf((character) value);
               
               if (len <= 0)
                  len = 1;
               
               fmt = "x(" + len + ")"; 
            }
            
            if (fmt == null)
            {
               displayFormat = DisplayFormatFactory.makeDefaultFormat(value);
            }
            else
            {
               displayFormat = DisplayFormatFactory.makeFormat(value, fmt);
            }
            
            displayDataPres = displayFormat.fromVar(displayVar);
         }
         
         catch (Exception e)
         {
            // do not throw other exceptions here, just ignore it
            e.printStackTrace(System.err);
            displayDataPres = null;
         }
         
         return;
      }

      // non-override path
      clearOverride();
      
      appVar = value.duplicate();
      
      // remember the size before it is possibly changed (updateSize() and
      // setAttrFormat() both can change the size)
      Dimension saved = dimension();
      
      updateSize();

      if (appVar.isUnknown() &&
          isEnabled() &&
          !ThinClient.getInstance().isInEnable())
      {
         setDrawDateSeparators(false);
      }

      if (autoFormat)
      {
         try
         {
            // note that order of evaluation is important (setAttrFormat
            // may change the size so getHeight() and getWidth() may differ
            // from the saved versions)
            if (setAttrFormat(null) &&
                (saved.height != height() || saved.width != width()))
            {
               Widget frame = UiUtils.locateFrame(this);
               
               if (frame instanceof Frame)
                  ((Frame) frame).relayout();
            }
         }
         catch (DisplayFormatParsingException e)
         {
            // can't happen because an automatically generated 
            // format string is always valid
            throw new RuntimeException("Format parsing error", e);
         }
      }
      else
      {
         display();
      }
      
      Label label = getLabelInstance();
      
      if (label != null)
         label.setEmptyModeOff();
   }

   /**
    * Get status of @ field flag.
    * 
    * @return  Current value of the flag.
    */
   public boolean isAtField()
   {
      return config.isAtField();
   }
   
   /**
    * Set widget width via configuration.
    * 
    * @param   width
    *          New value of the attribute.
    */
   public void setWidthChars(int width)
   {
      config.setWidthChars(width);
   }
   
   /**
    * Control how to handle last row of the empty space during drawing. This
    * is effective only in header mode.
    * 
    * @param   drawLine
    *          If this parameter is <code>true</code> then last row of the
    *          space occupied by widget is drawn as horizontal line.
    * @param   deltaWidth
    *          Extra horizontal space which will be occupied by widget in
    *          header mode.
    */
   public void setHeaderParm(boolean drawLine, int deltaWidth)
   {
      if (!config.isHeader())
         return;

      this.drawLine = drawLine;
      this.deltaWidth = deltaWidth;
   }

   /**
    * Get current state of the flag responsible for drawing line.
    * 
    * @return  Current state of the flag.
    */
   public boolean isDrawLine()
   {
      return drawLine;
   }
   
   /**
    * Get widget height.
    * 
    * @return  Widget height.
    */
   @Override
   public int height()
   {
      return super.height() + ((config.isHeader() && drawLine) ? 1 : 0);
   }

   /**
    * Get widget width.
    * 
    * @return  Widget width.
    */
   @Override
   public int width()
   {
      return super.width() + ((config.isHeader()) ? deltaWidth : 0);
   }

   /**
    * 
    * @see java.lang.Object#toString()
    */
   @Override
   public String toString()
   {
      return "FillIn " + location() + " " + dimension() +
            " label = " + getLabelText() + " value = " + getVariable() +
            " [" + getText(true) + "]";
   }

   /**
    * Turn on/off text group (word wrap) mode.
    * 
    * @param   enable
    *          New state of the word wrap mode.
    */
   public void setTextGroup(boolean enable)
   {
      groupMode = enable;
   }
   
   /**
    * Clear any temporary display override so that future drawing will be 
    * based upon our standard data and format.
    */
   private void clearOverride()
   {
      displayVar      = null;
      displayFormat   = null;
      displayDataPres = null;
   }

   /**
    * Synchronizes the actual size with the default size.
    */
   private void updateSize()
   {
      if (config.isAutoResize())
      {
         config.setWidthChars(appFormat.getScreenWidth());
         
         if (config.getHeightChars() <= 0)
            config.setHeightChars(1);

         config.setAutoResize(true);
      }
   }

   /**
    * Set type of the bound variable.
    * 
    * @param   dataType
    *          New data type name.
    */
   public void setDataType(String dataType)
   {
      if (dataType == null || 
         (appVar != null && 
          appVar.getTypeName().equalsIgnoreCase(config.getDataType())))
      {
         return;
      }

      appVar = DisplayFormat.instanceOfType(dataType);
      
      try
      {
         if (config.getFormat() != null)
            appFormat = DisplayFormatFactory.makeFormat(appVar, config.getFormat());
         else
            appFormat = DisplayFormatFactory.makeDefaultFormat(appVar);

         config.setDataType(appVar.getTypeName());
      }
      catch (DisplayFormatParsingException e)
      {
         throw new RuntimeException("Unable to set data type", e);
      }
      
      setCursorOffset(0);
      updateSize();
      display();
   }

   /**
    * Link editor with the <code>Browse</code> widget.
    * 
    * @param   browse
    *          The instance of the <code>Browse</code> to link with.
    */
   public void setBrowse(Browse browse)
   {
      this.browse = browse;
   }

   /**
    * Determines whether this {@link FillIn} has an associated {@link Browse}
    * widget.
    * 
    * @return   <code>true</code> if this {@link FillIn} has an associated
    *           {@link Browse} widget.
    */
   public boolean isLinkedToBrowse()
   {
      return browse != null;
   }

   /**
    * Determines whether auto-zap mode is enabled at this moment.
    * 
    * @return   <code>true</code> if auto-zap mode is enabled.
    */
   public boolean isZapActive()
   {
      // TODO: the groupMode exclusion below has not been confirmed in a 4GL
      //       environment, but it is known to solve at least one deviation
      //       where a text group is used on a set of widgets created from a
      //       character array; this usage makes the widgets operate in a mode
      //       where the contents of each widget don't clear when the user
      //       starts typing;
      return justActivated  &&
             !pendingNoZap &&
             config.isAutoZap() &&
             !groupMode &&          // this is the potentially incorrect change
             appFormat.isZapSupported();
   }

   /**
    * Reset the zap mode.
    */
   public void setPendingNoZap()
   {
      pendingNoZap = true;
   }

   /**
    * Get reference to last widget in the TEXT group.
    * 
    * @return  The reference to last widget in group.
    */
   FillIn getLastInGroup()
   {
      if (!groupMode)
         return this;
      
      FillIn ptr = this;
      
      do
      {
         if (ptr.isLastInGroup())
            break;
         
         ptr = ptr.getNextInGroup();
      }
      while(true);
      
      return ptr;
   }

   /**
    * Get reference to first widget in TEXT group.
    * 
    * @return  The reference to first widget in group.
    */
   public FillIn getFirstInGroup()
   {
      if (!groupMode)
         return this;
      
      FillIn ptr = this;
      
      do
      {
         if (ptr.isFirstInGroup())
            break;
         
         ptr = ptr.getPrevInGroup();
      }
      while(true);
      
      return ptr;
   }
   
   /**
    * Return <code>true</code> if last {@link #draw(Color)} was performed
    * with colors other than regular.
    * 
    * @return current state of selected flag.
    */
   public boolean isSelected()
   {
      return selected;
   }

   /**
    * Set format check flag.
    * 
    * @param    formatCheck
    */
   public void setFormatCheck(boolean formatCheck)
   {
      this.formatCheck  = formatCheck; 
   }
   
   /**
    * Get current state of hard format check flag.
    * 
    * @return   current state of the flag.
    */
   public boolean isFormatCheck()
   {
      return formatCheck || getVariable() instanceof date;
   }

   /**
    * Skip next activation request.
    * 
    * @param    source
    *           Source widget which looses the focus.
    * @param    newPos
    *           New cursor position.          
    */
   public void silentActivation(Widget source, int newPos)
   {
      skipActivation = true;
      
      if (source instanceof FillIn)
      {
         FillIn fin = (FillIn) source;

         if (groupMode && fin.groupMode && partOfGroup(fin))
            internalActivation(fin);
      }
      
      if (newPos != 0 || appDataPres == null)
         setCursorOffset(newPos);
      else
         setCursorOffset(appDataPres.getInitialCursorPos());
   }

   /**
    * Get valid cursor position.
    * 
    * @return   valid cursor position.
    */
   private int getValidCursorPos()
   {
      int res = getCursorOffset() + (isRightAlignedInt() ? leftPad : 0);
      
      if (res >= width())
         res--;
      
      return res;
   }
   
   /**
    * Completes edit and checks the result. Shows an error message if
    * necessary.
    * 
    * @return  Does edit successfully completed.
    */
   private boolean completeEdit()
   {
      if (appDataPres != null)
      {
         if (appDataPres.finishEdit())
         {
            // add repainting if necessary 
         }
      }

      return true;
   }

   /**
    * Returns the insert mode state.
    * 
    * @return  The insert mode state.
    *          <ul>
    *             <li><code>false</code>- overwrite mode,
    *             <li><code>true</code>- insert mode.
    *          </ul>
    */
   private boolean getInsertMode()
   {
      if (appDataPres != null)
      {
         return appDataPres.getInsertMode();
      }

      return Window.getInsertMode() == Window.INSERT_MODE_INS;
   }

   /**
    * Changes the insert mode state in the window.
    * 
    * @param   mode
    *          New state.
    *          <ul>
    *             <li><code>false</code>- overwrite mode,
    *             <li><code>true</code>- insert mode.
    *          </ul>
    */
   private void setWindowInsertMode(boolean mode)
   {
      Window.setInsertMode((mode) ? Window.INSERT_MODE_INS : 
                                    Window.INSERT_MODE_OVR);
   }

   /**
    * Refill content after insertion or deletion of characters.
    * 
    * @param   insert
    *          Mode of operation - <code>true</code> means that character
    *          was inserted.
    * @param   key
    *          The character which will be inserted.
    * 
    * @return  <code>true</code> if after insertion operation cursor is
    *          allowed to go to the next widget.
    */
   private boolean refill(boolean insert, int key)
   {
      if (isLastInGroup())
         return false; 

      if (insert)
      {
         // character is about to be inserted

         String text = getText();
         int len = text.length();
         boolean insertMode = getInsertMode();
         
         text = StringHelper.safeTrimTrailing(text);
         
         if (!insertMode && key != ' ')
         {
            // override mode quirk: 
            // entering last character in the field is handled like insertion 
            
            if (text.length() > getCursorOffset())
               return true;
         }

         if (getCursorOffset() < (len - 1) && (len - 1) > text.length())   
         {
            // there is enough space
            return true;
         }
         
         String word = getLastWord(-1);
         
         if (word.length() == len || 
             word.length() == 0)  // last word is too long or too short
         {
            if (word.length() == len || getCursorOffset() < len)
               return false;
            
            //entering character into last position always moves cursor
            //to the next row and handled like insertion  
         }

         // in non-insert mode, if the key is ' ', the last word will never be 
         // wrapped to the next line, as line size doesn't change; if the char
         // is not blank, then the last word in the line is wrapped to the 
         // next line
         if ((!insertMode && getCursorOffset() < len) || 
             (key == ' '  && getCursorOffset() == len))
         {
            word = "";
         }
         
         int new_len = len - word.length();
         int delta = getCursorOffset() - new_len;
         
         text = getText().substring(0, new_len);
         
         setValueInt(new character(text));

         FillIn ptr = getNextInGroup();
         
         // only insert mode or typing a char on the very last position in the 
         // line allows word wrapping to the next line
         if (insertMode || (!insertMode && getCursorOffset() == len))
         {
            if (word.length() != 0 || key != ' ')
               ptr.insertWord(word, key != ' ' && delta > 0);
         }
         
         if (delta >= 0)
         {
            // move cursor to the next row. this doesn't depend on the insert
            // mode - in both cases, cursor is moved if the condition is met
            ptr.internalActivation(this);

            FocusManager.tryFocusChange(ptr);
            
            ptr.setCursorOffset(delta);
            
            if (delta > 0 || key != ' ')
            {
               EventManager.postEvent(new KeyInput(ptr,
                                              EventType.KEY_TYPED, 
                                              key));
               keyApplied = true;
            }
         }
         
         return true;
      }
      
      // character was removed
      String text = getText();
      int len = text.length(); // for trailing space

      text = StringHelper.safeTrimTrailing(text);

      len -= text.length();

      FillIn ptr = getNextInGroup();

      String words = ptr.getLeadingWords(len);

      if (text.length() != 0)
      {
         // deletion of the trailing blanks does not trigger
         // refill
         if (key == ' ' && getCursorOffset() >= text.length())
            return true;

         if (ptr.getNumLeadingSpaces() > getCursorOffset())
            return true;
      }

      if (words.length() > 0)
      {
         String line = text + (text.length() == 0 ? "" : " ") + words;
         setValueInt(new character(line));

         text = ptr.getText().substring(words.length() + 1);
         ptr.setValueInt(new character(text));
         ptr.setCursorOffset(0);

         ptr.refill(false, ' ');
      }

      
      return true;
   }

   /**
    * Get number of leading spaces in the text before first non-space
    * character.
    * 
    * @return  The number of leading spaces.
    */
   private int getNumLeadingSpaces()
   {
      char[] text = getText().toCharArray();
      
      for (int i = 0; i < text.length; i++)
      {
         if (text[i] != ' ')
            return i - 1;
      }
      
      return text.length;
   }

   /**
    * Set value and preserve state of the insert mode and cursor offset.
    * 
    * @param   value
    *          New value for the fill-in.
    */
   private void setValueInt(BaseDataType value)
   {
      boolean save = getInsertMode();
      int offset = getCursorOffset();
      
      setValue(value);
      setInsertMode(save);
      setCursorOffset(offset);
      config.setModified(true);
   }
   
   /**
    * Prepares fill-in for internal activation inside TEXT group.
    * 
    * @param   source
    *          The widget which is about to trigger activation.
    */
   private void internalActivation(FillIn source)
   {
      skipActivation  = true;

      // In the group mode, do not refresh the status string when we are
      // moving between two fill-ins with undefined help string.
      skipHelpAdjustment = (config.getHelp() == null) &&
                           (source.config.getHelp() == null);
      
      setInsertMode(source.getInsertMode());
      setCursorOffset(source.getCursorOffset());
   }
   
   /**
    * Insert word at the beginning of the existing text. If there is not
    * enough room to put provided word, then few trailing words are shifted
    * into next widgets in the TEXT group.
    * 
    * @param   word
    *          The word to be inserted.
    *          
    * @param   override
    *          Mark special case of override mode.
    */
   private void insertWord(String word, boolean override)
   {
      FillIn ptr = this;
      
      while (ptr != null)
      {
         String text = ptr.getText();
         int len = text.length();
         
         word = StringHelper.safeTrimTrailing(word);
         
         if (override)
            word += " ";

         text = StringHelper.safeTrimTrailing(text);

         if ((text.length() + word.length() + 1) <= len)
         {
            ptr.setValueInt(new character(word + 
                                          ((override) ? "" : " ") + 
                                          text));
            break;
         }

         String trailing = ptr.getLastWord(word.length() + 1);
         
         text = ptr.getText().substring(0, len - trailing.length());
         
         text = word + 
               ((override) ? "" : " ") + 
               StringHelper.safeTrimTrailing(text);
         
         ptr.setValueInt(new character(text));
         
         word = trailing;
         ptr = ptr.getNextInGroup();
      }
   }

   /**
    * Retrieve last word(s) from the buffer. Total length of the retrieved
    * text should be at least of specified length. If limit is less than
    * <code>0</code> then last word is retrieved.
    * 
    * @param   limit
    *          The minimum length of the retrieved text.
    * 
    * @return  The retrieved word(s).
    */
   private String getLastWord(int limit)
   {
      char[] text = getText().toCharArray();
      
      int offset = text.length;
      limit = text.length - limit;
      
      for (int i = text.length - 1; i >= 0; i--)
      {
         if (text[i] != ' ')
         {
            offset = i;
            continue;
         }

         if (i > limit)
            continue;
         
         break;
      }
      
      if (text.length - offset <= 0)
         return "";
      
      return new String(text, offset, text.length - offset);
   }

   /**
    * Get leading words from the text buffer.
    * 
    * @param   size
    *          Total size of the words should be less than specified size.
    *          
    * @return  The leading words.
    */
   private String getLeadingWords(int size)
   {
      char[] text = getText().toCharArray();
      
      int lastSpace = -1;
      int lastNonSpace = -1;
      
      for (int i = 0; i < size; i++)
      {
         if (text[i] == ' ')
            lastSpace = i;
         else
            lastNonSpace = i;
      }
      
      // there is no word or word is longer than specified size
      if (lastNonSpace < 0 || lastSpace < 0)
         return "";
      
      return StringHelper.safeTrimTrailing(new String(text, 0, lastSpace));
   }

   /**
    * Check if given widget is part of the same group as this widget.
    * 
    * @param   comp
    *          The widget to check.
    * 
    * @return  <code>true</code> if widget belongs to the same group.
    */
   private boolean partOfGroup(Widget comp)
   {
      if (!groupMode)
         return false;
      
      FillIn ptr = getFirstInGroup();
      
      while (ptr != null)
      {
         if (ptr == comp)
            return true;
         
         ptr = ptr.getNextInGroup();
      }
      
      return false;
   }
   
   /**
    * Get previous component in TEXT clause group.
    *  
    * @return  The previous component in group. 
    */
   private FillIn getPrevInGroup()
   {
      int id = config.getTextGroupPrev();
      
      return (FillIn) ThinClient.getInstance().getWidget(id); 
   }

   /**
    * Get next component in TEXT clause group.
    * 
    * @return  The next component in group.
    */
   private FillIn getNextInGroup()
   {
      int id = config.getTextGroupNext();
      
      return (FillIn) ThinClient.getInstance().getWidget(id); 
   }

   /**
    * Check if component is first in the TEXT clause group.
    *  
    * @return  <code>true</code> if component is first in group.
    */
   private boolean isFirstInGroup()
   {
      return (getPrevInGroup() == null);
   }

   /**
    * Check if component is last in the TEXT clause group.
    *  
    * @return  <code>true</code> if component is last in group.
    */
   private boolean isLastInGroup()
   {
      if (!groupMode)
         return false;
      
      return (getNextInGroup() == null);
   }

   /**
    * Get focus traversable widget which is right before the TEXT group in TAB
    * order.
    * 
    * @return  The widget which is before TEXT group.
    */
   private Widget getPrevBeforeGroup()
   {
      FillIn first = getFirstInGroup();
      FillIn last = getLastInGroup();
      
      Widget[] list = parent().widgets();
      
      int start = -1;
      int end = -1;
      
      for (int i = 0; i < list.length; i++)
      {
         if (list[i] == first)
            start = i; 
         if (list[i] == last)
            end = i; 
      }
      
      if (start < 0 || end < 0)
         return first; // first in group
      
      for (int i = start - 1; i >= 0; i--)
      {
         if (list[i].focusTraversable())
            return list[i];
      }
      
      for (int i = list.length - 1; i > end; i--)
      {
         if (list[i].focusTraversable())
            return list[i];
      }
      
      return first; // first in group
   }

   /**
    * Get focus traversable widget which is right before the TEXT group in TAB
    * order.
    * 
    * @return  The widget which is before TEXT group.
    */
   private Widget getNextAfterGroup()
   {
      FillIn first = getFirstInGroup();
      FillIn last = getLastInGroup();
      
      Widget[] list = parent().widgets();
      
      int start = -1;
      int end = -1;
      
      for (int i = 0; i < list.length; i++)
      {
         if (list[i] == first)
            start = i; 
         if (list[i] == last)
            end = i; 
      }
      
      if (start < 0 || end < 0)
         return first; // first in group
      
      for (int i = end + 1; i < list.length; i++)
      {
         if (list[i].focusTraversable())
            return list[i];
      }
      
      for (int i = 0; i < start; i++)
      {
         if (list[i].focusTraversable())
            return list[i];
      }
      
      return first; // first in group
   }

   /**
    * If fill-in displays a date value, it forces separators for the unknown
    * date value to be drawn or not in the subsequent value rendering (applies
    * to editing mode only).
    *
    * @param drawSeparators
    *        If <code>true</code> then separators will be displayed for the
    *        unknown date value.
    */
   private void setDrawDateSeparators(boolean drawSeparators)
   {
      if (appFormat != null && appFormat instanceof DateFormat)
      {
         ((DateFormat) appFormat).setDrawSeparators(drawSeparators);
      }
   }

   /**
    * Some actions on focus hanging.
    */
   private class LocalFocusListener implements FocusListener
   {
      /**
       * Setups <code>FillIn</code> as it is just activated by Tab key.
       * 
       * @param   e
       *          FocusEvent instance (ignored).
       */
      @Override
      public void onFocusGained(FocusEvent e)
      {
         if (e == null)
            return;
       
         widgetFocused = true;

         if (!skipHelpAdjustment)
            ThinClient.setHelp(config.getHelp());

         // FOCUS_GAINED events must not change activated state
         activate(false);
         
         /*
         // this activation fix is backed-out as it exposes latent issues 
         // about the ENTRY event processing
         activate(true);
         */
      }

      /**
       * Reset SELECTED attribute when focus is lost.
       * 
       * @param   e
       *          FocusEvent instance (ignored).
       */
      @Override
      public void onFocusLost(FocusEvent e)
      {
         if (e == null)
            return;
         
         widgetFocused = false;
      }
   }
}
